#define APPLICATION TEXT("OGMcleaver")
#define APP_VERSION APPLICATION TEXT(" 1.1.0")

#include <windows.h>
#include <shlobj.h>
#include <stdio.h>

#include "safefgets/safefgets.h"
#include "winutils/winutils.h"

#include "resource.h"

#define MAX_LOADSTRING 1024

HINSTANCE hInstance;

struct {
	TCHAR avc2avi   [MAX_PATH];
	TCHAR OGMInfo   [MAX_PATH]; 
	TCHAR OGMDemuxer[MAX_PATH];
	TCHAR MKVToolnix[MAX_PATH];
	TCHAR MKVInfo   [MAX_PATH];
	TCHAR MKVExtract[MAX_PATH];
} Tools;

VOID InitializeTools()
{
	TCHAR module[MAX_PATH];
	TCHAR path[MAX_PATH];
	LPTSTR fname;

	GetModuleFileName(NULL, module, MAX_PATH);
	GetFullPathName(module, MAX_PATH, path, & fname);
	(*fname) = 0;

	lstrcpy(Tools.avc2avi,    path);
	lstrcat(Tools.avc2avi,    TEXT("avc2avi\\avc2avi.exe"));
	lstrcpy(Tools.OGMInfo,    path);
	lstrcat(Tools.OGMInfo,    TEXT("ogmtools\\OGMInfo.exe"));
	lstrcpy(Tools.OGMDemuxer, path);
	lstrcat(Tools.OGMDemuxer, TEXT("ogmtools\\OGMDemuxer.exe"));
	
	*path = 0;
	if (GetFullPathInSpecialFolder(CSIDL_APPDATA, TEXT("OGMcleaver"), NULL, TEXT("MKVToolnix"), module, MAX_PATH, FALSE, NULL)) {
		char filename[MAX_PATH];
		FILE * f;
		#ifdef UNICODE
		WideCharToMultiByte(CP_ACP, 0, module, -1, filename, MAX_PATH, NULL, NULL);
		#else
		lstrcpy(filename, module);
		#endif
		f = fopen(filename, "rt");
		if (f) {
			char * line = NULL;
			line = safefgets(line, f);
			if (line) {
				#ifdef UNICODE
				MultiByteToWideChar(CP_ACP, 0, line, -1, path, MAX_PATH);
				#else
				lstrcpyn(path, line, MAX_PATH);
				#endif
				free(line);
			}
			fclose(f);
		}
	}
	if (! *path) {
		GetFullPathInSpecialFolder(CSIDL_PROGRAM_FILES, NULL, NULL, TEXT("MKVToolnix"), path, MAX_PATH, FALSE, NULL);
	}
	lstrcpy(Tools.MKVToolnix, path);
	lstrcpy(Tools.MKVInfo,    path);
	lstrcat(Tools.MKVInfo,    TEXT("\\MKVInfo.exe"));
	lstrcpy(Tools.MKVExtract, path);
	lstrcat(Tools.MKVExtract, TEXT("\\MKVExtract.exe"));
}

#define VIDEO_TRACK 0
#define AUDIO_TRACK 1
#define SUBTITLE_TRACK 2
#define TRACK_TYPES 3
#define UNKNOWN_TRACK 3

struct tagOGMTRACK
{
	int Type;
	int Index;
	int StreamID;
};
typedef struct tagOGMTRACK OGMTRACK;
typedef struct tagOGMTRACK * LPOGMTRACK;

struct tagMKVTRACK
{
	int Type;
	int Index;
	int TrackNo;
	TCHAR Extension[6];
	double FrameRate;
	TCHAR ExtractFile[MAX_PATH];
	TCHAR ConvertFile[MAX_PATH];
};
typedef struct tagMKVTRACK MKVTRACK;
typedef struct tagMKVTRACK * LPMKVTRACK;

struct tagINPUTFILE
{
	TCHAR FileName[MAX_PATH];
	BOOL IsOGM;
	int TrackCount;
	int TypeCount[TRACK_TYPES];
	LPOGMTRACK OGMTrack;
	LPMKVTRACK MKVTrack;
};
typedef struct tagINPUTFILE INPUTFILE;
typedef struct tagINPUTFILE * LPINPUTFILE;

typedef BOOL TRACKARRAY[TRACK_TYPES];
typedef TRACKARRAY * LPTRACKARRAY;
struct tagTRACKMATRIX
{
	LPTRACKARRAY Type;
	int Count;
};
typedef struct tagTRACKMATRIX TRACKMATRIX;
typedef struct tagTRACKMATRIX * LPTRACKMATRIX;

LPINPUTFILE ParseOGMInfo(LPCTSTR FileName)
{
	
	LPINPUTFILE InputFile;
	char filename[MAX_PATH];
	FILE * f;
	char * line = NULL;
	int streamid, dummy;
	char * type = NULL;
	int Count[4] = {0, 0, 0, 0};
	
	#ifdef UNICODE
	WideCharToMultiByte(CP_ACP, 0, FileName, -1, filename, MAX_PATH, NULL, NULL);
	#else
	lstrcpy(filename, FileName);
	#endif
	f = fopen(filename, "rt");
	if (! f) return NULL;
	InputFile = (LPINPUTFILE) FatalAlloc(sizeof(INPUTFILE));
	InputFile->IsOGM = TRUE;
	while ((line = safefgets(line, f)) != NULL) {
		type = (char *) FatalReAlloc(type, sizeof(char) * strlen(line));
		if ((sscanf(line, "=== STREAM %d ( %d %s", & streamid, & dummy, type) == 3) &&
			(type[strlen(type) - 1] == ')')) {
			InputFile->OGMTrack = (LPOGMTRACK) FatalReAlloc(InputFile->OGMTrack, (InputFile->TrackCount + 1) * sizeof(OGMTRACK));
			InputFile->OGMTrack[InputFile->TrackCount].StreamID = streamid;
			if (! strcmp(type, "VIDEO)")) {
				InputFile->OGMTrack[InputFile->TrackCount].Type  = VIDEO_TRACK;
				InputFile->OGMTrack[InputFile->TrackCount].Index = ++(InputFile->TypeCount[VIDEO_TRACK]);
			} else if (! strcmp(type, "AUDIO)") || ! strcmp(type, "VORBIS)")) {
				InputFile->OGMTrack[InputFile->TrackCount].Type  = AUDIO_TRACK;
				InputFile->OGMTrack[InputFile->TrackCount].Index = ++(InputFile->TypeCount[AUDIO_TRACK]);
			} else if (! strcmp(type, "TEXT)") || ! strcmp(type, "TEXT)")) {
				InputFile->OGMTrack[InputFile->TrackCount].Type  = SUBTITLE_TRACK;
				InputFile->OGMTrack[InputFile->TrackCount].Index = ++(InputFile->TypeCount[SUBTITLE_TRACK]);
			} else {
				InputFile->OGMTrack[InputFile->TrackCount].Type  = UNKNOWN_TRACK;
			}
			InputFile->TrackCount++;
		}
	}
	FatalFree(type);
	fclose(f);
	if (! InputFile->TrackCount) {
		free(InputFile);
		InputFile = NULL;
	}
	return InputFile;
}

#define Is(type, name) (! strncmp((type), (name), strlen(name)))

LPINPUTFILE ParseMKVInfo(LPCTSTR FileName)
{
	LPINPUTFILE InputFile;
	char filename[MAX_PATH];
	FILE * f;
	char * line = NULL;
	int trackno;
	float ms, fps;
	char * type = NULL;
	LPTSTR extension;
	
	#ifdef UNICODE
	WideCharToMultiByte(CP_ACP, 0, FileName, -1, filename, MAX_PATH, NULL, NULL);
	#else
	lstrcpy(filename, FileName);
	#endif
	f = fopen(filename, "rt");
	if (! f) return NULL;
	InputFile = (LPINPUTFILE) FatalAlloc(sizeof(INPUTFILE));
	InputFile->IsOGM = FALSE;
	while ((line = safefgets(line, f)) != NULL) {
		if (! strcmp(line, "| + A track\n")) {
			InputFile->MKVTrack = (LPMKVTRACK) FatalReAlloc(InputFile->MKVTrack, (InputFile->TrackCount + 1) * sizeof(MKVTRACK));
			InputFile->MKVTrack[InputFile->TrackCount].TrackNo = -1;
			InputFile->MKVTrack[InputFile->TrackCount].Type    = UNKNOWN_TRACK;
			lstrcpy(InputFile->MKVTrack[InputFile->TrackCount].Extension, TEXT(""));
			lstrcpy(InputFile->MKVTrack[InputFile->TrackCount].ConvertFile, TEXT(""));
			InputFile->TrackCount++;
		} else if (InputFile->TrackCount) {
			type = (char *) FatalReAlloc(type, sizeof(char) * strlen(line));
			if (sscanf(line, "|  + Track number: %d", & trackno) == 1) {
				InputFile->MKVTrack[InputFile->TrackCount - 1].TrackNo = trackno;
			} else if (sscanf(line, "|  + Track type: %s", type) == 1) {
				if (! strcmp(type, "video")) {
					InputFile->MKVTrack[InputFile->TrackCount - 1].Type  = VIDEO_TRACK;
					InputFile->MKVTrack[InputFile->TrackCount - 1].Index = ++(InputFile->TypeCount[VIDEO_TRACK]);
				} else if (! strcmp(type, "audio")) {
					InputFile->MKVTrack[InputFile->TrackCount - 1].Type  = AUDIO_TRACK;
					InputFile->MKVTrack[InputFile->TrackCount - 1].Index = ++(InputFile->TypeCount[AUDIO_TRACK]);
				} else if (! strcmp(type, "subtitles")) {
					InputFile->MKVTrack[InputFile->TrackCount - 1].Type  = SUBTITLE_TRACK;
					InputFile->MKVTrack[InputFile->TrackCount - 1].Index = ++(InputFile->TypeCount[SUBTITLE_TRACK]);
				}
			} else if (sscanf(line, "|  + Codec ID: %s", type) == 1) {
				if (Is(type, "V_MPEG1") || Is(type, "V_MPEG2")) {
					extension = TEXT(".mpg");
				} else if (Is(type, "V_MS/VFW/FOURCC")) {
					extension = TEXT(".avi");
				} else if (Is(type, "V_MPEG4/ISO/AVC")) {
					extension = TEXT(".h264");
				} else if (Is(type, "A_MPEG/L3")) {
					extension = TEXT(".mp3");
				} else if (Is(type, "A_MPEG/L2")) {
					extension = TEXT(".mp2");
				} else if (Is(type, "A_MPEG/L1")) {
					extension = TEXT(".mpa");
				} else if (Is(type, "A_AC3")) {
					extension = TEXT(".ac3");
				} else if (Is(type, "A_DTS")) {
					extension = TEXT(".dts");
				} else if (Is(type, "A_PCM/INT/LIT")) {
					extension = TEXT(".wav");
				} else if (Is(type, "A_AAC")) {
					extension = TEXT(".aac");
				} else if (Is(type, "A_QUICKTIME")) {
					extension = TEXT(".qdm");
				} else if (Is(type, "A_TTA1")) {
					extension = TEXT(".tta");
				} else if (Is(type, "A_WAVPACK4")) {
					extension = TEXT(".wv");
				} else if (Is(type, "A_VORBIS")) {
					extension = TEXT(".ogg");
				} else if (Is(type, "S_TEXT/UTF8")) {
					extension = TEXT(".srt");
				} else if (Is(type, "S_TEXT/SSA")) {
					extension = TEXT(".ssa");
				} else if (Is(type, "S_TEXT/ASS")) {
					extension = TEXT(".ass");
				} else if (Is(type, "S_TEXT/USF")) {
					extension = TEXT(".usf");
				} else if (Is(type, "S_VOBSUB")) {
					extension = TEXT(".idx");
				} else if (Is(type, "A_REAL")) {
					extension = TEXT(".ra");
				} else if (Is(type, "V_REAL")) {
					extension = TEXT(".rmvb");
				} else {
					extension = TEXT("");
				}
				lstrcpy(InputFile->MKVTrack[InputFile->TrackCount - 1].Extension, extension);
			} else if (sscanf(line, "|  + Default duration: %fms (%f", & ms, & fps) == 2) {
				InputFile->MKVTrack[InputFile->TrackCount - 1].FrameRate = fps;
			}
		}
	}
	free(type);
	fclose(f);
	if (! InputFile->TrackCount) {
		free(InputFile);
		InputFile = NULL;
	}
	return InputFile;
}

LPINPUTFILE AnalyzeFile(LPCTSTR FileName)
{
	LPINPUTFILE InputFile = NULL;
	TCHAR Command[MAX_PATH * 2];
	int pos;
	TCHAR TempFileName[MAX_PATH];
	TCHAR path[MAX_PATH - 12];
	GetTempPath(MAX_PATH - 12, path);
	GetTempFileName(path, TEXT("OCL"), 0, TempFileName);
	pos = lstrlen(FileName);
	while (pos && FileName[pos] != TEXT('.')) pos--;
	if (! lstrcmpi(FileName + pos, TEXT(".OGM")) || ! lstrcmpi(FileName + pos, TEXT(".OGG"))) {
		wsprintf(Command, TEXT("--headers \"%s\""), FileName);
		if (ExecuteCommand(Tools.OGMInfo, Command, NULL, TempFileName, NULL, NULL, NULL)) {
			InputFile = ParseOGMInfo(TempFileName);
		}
	} else if (! lstrcmpi(FileName + pos, TEXT(".MKV"))) {
		wsprintf(Command, TEXT("\"%s\""), FileName);
		if (ExecuteCommand(Tools.MKVInfo, Command, NULL, TempFileName, NULL, NULL, NULL)) {
			InputFile = ParseMKVInfo(TempFileName);
		}
	}
	if (InputFile) {
		lstrcpy(InputFile->FileName, FileName);
	}
	DeleteFile(TempFileName);
	return InputFile;
}

BOOL ExtractOGMTracks(LPINPUTFILE InputFile, BOOL TrackFirst, BOOL NameTrackType, BOOL ConvertH264, LPTRACKMATRIX TrackMatrix, LPBOOL IsCancelled)
{
	BOOL cmdOk;
	BOOL allOk = TRUE;
	int i;
	TCHAR command[MAX_PATH * 3]; 
	wsprintf(command, TEXT("tracks \"%s\" --output-path ."), InputFile->FileName);
	cmdOk = ExecuteCommand(Tools.OGMDemuxer, command, NULL, NULL, NULL, IsCancelled, NULL);
	for (i = 0; i < InputFile->TrackCount; i++) {
		BOOL ok = FALSE;
		TCHAR fileMask[MAX_PATH]; 
		TCHAR findMask[MAX_PATH];
		int pos;
		HANDLE findHandle;
		WIN32_FIND_DATA findData;
		for (pos = lstrlen(InputFile->FileName);
			pos && InputFile->FileName[pos - 1] != TEXT('\\');
			pos--);
		lstrcpy(fileMask, InputFile->FileName + pos);
		lstrcat(fileMask, TEXT(".%d.*"));
		wsprintf(findMask, fileMask, InputFile->OGMTrack[i].StreamID);
		findHandle = FindFirstFile(findMask, & findData);
		if (findHandle != INVALID_HANDLE_VALUE) {
			if (cmdOk && (InputFile->OGMTrack[i].Type != UNKNOWN_TRACK) && TrackMatrix->Type[InputFile->OGMTrack[i].Index - 1][InputFile->OGMTrack[i].Type]) {
				TCHAR extension[MAX_PATH];
				TCHAR fileName[MAX_PATH];
				TCHAR outputFile[MAX_PATH];
				LPTSTR Prefix;
				for (pos = lstrlen(findData.cFileName);
					pos && findData.cFileName[pos] != TEXT('.');
					pos--);
				lstrcpy(extension, findData.cFileName + pos);
				for (pos--; pos && findData.cFileName[pos] != '.'; pos--);
				if (pos > 3 && 
					(findData.cFileName[pos - 3] == 'o' || findData.cFileName[pos - 3] == 'O') && 
					(findData.cFileName[pos - 2] == 'g' || findData.cFileName[pos - 2] == 'G') && 
					(findData.cFileName[pos - 1] == 'm' || findData.cFileName[pos - 1] == 'M') && 
					findData.cFileName[pos] == '.') pos -= 4;
				lstrcpy(fileName, findData.cFileName);
				fileName[pos] = 0;
				if (! NameTrackType) {
					Prefix = TEXT("Track");
				} else if (InputFile->OGMTrack[i].Type == VIDEO_TRACK) {
					Prefix = TEXT("Video");
				} else if (InputFile->OGMTrack[i].Type == AUDIO_TRACK) {
					Prefix = TEXT("Audio");
				} else if (InputFile->OGMTrack[i].Type == SUBTITLE_TRACK) {
					Prefix = TEXT("Subtitle");
				}
				if (TrackFirst) {
					wsprintf(outputFile, TEXT("%s%d-%s%s"), Prefix, InputFile->OGMTrack[i].Index, fileName, extension);
				} else {
					wsprintf(outputFile, TEXT("%s-%s%d%s"), fileName, Prefix, InputFile->OGMTrack[i].Index, extension);
				}
				ok = MoveFileEx(findData.cFileName, outputFile, MOVEFILE_REPLACE_EXISTING);
			} else {
				ok = DeleteFile(findData.cFileName);
			}
			FindClose(findHandle);
		}
		allOk = allOk && ok;
	}
	return cmdOk && allOk;
}

BOOL ExtractMKVTracks(LPINPUTFILE InputFile, BOOL TrackFirst, BOOL NameTrackType, BOOL ConvertH264, LPTRACKMATRIX TrackMatrix, LPBOOL IsCancelled)
{
	BOOL ok;
	int i;
	LPTSTR Command;
	TCHAR CmdPart[MAX_PATH * 4];

	TCHAR fileName[MAX_PATH];
	int pos;
	for (pos = lstrlen(InputFile->FileName);
		pos && InputFile->FileName[pos - 1] != TEXT('\\');
		pos--);
	lstrcpy(fileName, InputFile->FileName + pos);
	pos = lstrlen(fileName);
	if (pos > 4 && 
		(fileName[pos - 3] == TEXT('m') || fileName[pos - 3] == TEXT('M')) && 
		(fileName[pos - 2] == TEXT('k') || fileName[pos - 2] == TEXT('K')) && 
		(fileName[pos - 1] == TEXT('v') || fileName[pos - 1] == TEXT('V')) && 
		fileName[pos - 4] == TEXT('.')) pos -= 4;
	fileName[pos] = 0;
	
	wsprintf(CmdPart, TEXT("tracks \"%s\""), InputFile->FileName);
	Command = CopyStr(NULL, CmdPart);
	for (i = 0; i < InputFile->TrackCount; i++) {
		if (InputFile->MKVTrack[i].Type != UNKNOWN_TRACK) {
			if (TrackMatrix->Type[InputFile->MKVTrack[i].Index - 1][InputFile->MKVTrack[i].Type]) {
				LPTSTR Prefix;
				if (! NameTrackType) {
					Prefix = TEXT("Track");
				} else if (InputFile->MKVTrack[i].Type == VIDEO_TRACK) {
					Prefix = TEXT("Video");
				} else if (InputFile->MKVTrack[i].Type == AUDIO_TRACK) {
					Prefix = TEXT("Audio");
				} else if (InputFile->MKVTrack[i].Type == SUBTITLE_TRACK) {
					Prefix = TEXT("Subtitle");
				}
				if (TrackFirst) {
					wsprintf(InputFile->MKVTrack[i].ExtractFile, TEXT("%s%d-%s%s"), 
						Prefix, InputFile->MKVTrack[i].Index, fileName, InputFile->MKVTrack[i].Extension);
				} else {
					wsprintf(InputFile->MKVTrack[i].ExtractFile, TEXT("%s-%s%d%s"), 
						fileName, Prefix, InputFile->MKVTrack[i].Index, InputFile->MKVTrack[i].Extension);
				}
				wsprintf(CmdPart, TEXT(" %d:\"%s\""), InputFile->MKVTrack[i].TrackNo, InputFile->MKVTrack[i].ExtractFile);
				Command = ConcatStr(Command, CmdPart);
			}
		}
	}
	ok = ExecuteCommand(Tools.MKVExtract, Command, NULL, NULL, NULL, IsCancelled, NULL);
	for (i = 0; ok && i < InputFile->TrackCount; i++) {
		if (InputFile->MKVTrack[i].Type != UNKNOWN_TRACK) {
			if (TrackMatrix->Type[InputFile->MKVTrack[i].Index - 1][InputFile->MKVTrack[i].Type]) {
				if (ConvertH264 && InputFile->MKVTrack[i].Type == VIDEO_TRACK && ! lstrcmp(InputFile->MKVTrack[i].Extension, TEXT(".h264"))) {
					LPTSTR framerate;
					framerate = DoubleStr(InputFile->MKVTrack[i].FrameRate);
					lstrcpy(InputFile->MKVTrack[i].ConvertFile, InputFile->MKVTrack[i].ExtractFile);
					lstrcpy(InputFile->MKVTrack[i].ConvertFile + lstrlen(InputFile->MKVTrack[i].ExtractFile) - lstrlen(InputFile->MKVTrack[i].Extension), TEXT(".avi"));
					wsprintf(CmdPart, TEXT("-f %s -i \"%s\" -o \"%s\""), framerate, InputFile->MKVTrack[i].ExtractFile, InputFile->MKVTrack[i].ConvertFile);
					FatalFree(framerate);
					ok = ExecuteCommand(Tools.avc2avi, CmdPart, NULL, NULL, NULL, IsCancelled, NULL);
					if (ok) {
						ok = DeleteFile(InputFile->MKVTrack[i].ExtractFile);
						if (ok) lstrcpy(InputFile->MKVTrack[i].ExtractFile, TEXT(""));
					}
				}
			}
		}
	}
	if (! ok) {
		for (i = 0; i < InputFile->TrackCount; i++) {
			if (InputFile->MKVTrack[i].Type != UNKNOWN_TRACK) {
				if (TrackMatrix->Type[InputFile->MKVTrack[i].Index - 1][InputFile->MKVTrack[i].Type]) {
					if (InputFile->MKVTrack[i].ExtractFile[0]) DeleteFile(InputFile->MKVTrack[i].ExtractFile);
					if (InputFile->MKVTrack[i].ConvertFile[0]) DeleteFile(InputFile->MKVTrack[i].ConvertFile);
				}
			}
		}
	}
	
	return ok;
}

#define GetInputCount(hWnd) ((int) SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETCOUNT, 0, 0))

LPINPUTFILE GetInputFile(HWND hWnd, int Index)
{
	LONG ItemData;
	ItemData = SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETITEMDATA, Index, 0);
	if (ItemData == LB_ERR) return NULL;
	return (LPINPUTFILE) ItemData;
}

VOID AddInputFile(HWND hWnd, LPINPUTFILE InputFile)
{
	int Count;
	Count = GetInputCount(hWnd);
	SendDlgItemMessage(hWnd, IDC_FILELIST, LB_ADDSTRING, 0, (LPARAM) InputFile);
	// SendDlgItemMessage(hWnd, IDC_FILELIST, LB_SETITEMDATA, Count, (LPARAM) InputFile);
	SendDlgItemMessage(hWnd, IDC_FILELIST, LB_SETCURSEL, Count, 0);
	if (! SendDlgItemMessage(hWnd, IDC_OUTPUTDIRECTORY, WM_GETTEXTLENGTH, 0, 0)) {
		TCHAR Directory[MAX_PATH];
		LPTSTR DirEnd;
		if (GetFullPathName(InputFile->FileName, MAX_PATH, Directory, & DirEnd)) {
			*DirEnd = 0;
			SetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, Directory);
		}
	}
}

VOID RemoveInputFile(HWND hWnd, int Index)
{
	LPINPUTFILE InputFile;
	InputFile = GetInputFile(hWnd, Index);
	if (InputFile) {
		SendDlgItemMessage(hWnd, IDC_FILELIST, LB_DELETESTRING, Index, 0);
		if (InputFile->IsOGM) {
			FatalFree(InputFile->OGMTrack);
		} else {
			FatalFree(InputFile->MKVTrack);
		}
		FatalFree(InputFile);
		EnableWindow(GetDlgItem(hWnd, IDC_REMOVEFILES), SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETSELCOUNT, 0, 0));
	}
}

VOID RemoveSelectedInputFiles(HWND hWnd)
{
	int Count, i;
	Count = GetInputCount(hWnd);
	for (i = Count - 1; i >= 0; i--) {
		if (SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETSEL, i, 0)) {
			RemoveInputFile(hWnd, i);
		}
	}
}

VOID UpdateDialog(HWND hWnd)
{
	BOOL HasTrackToExtract = FALSE;
	int InputCount;
	InputCount = GetInputCount(hWnd);
	if (InputCount) {
		int TrackCount[TRACK_TYPES];
		int CaretIndex;
		int ListIndex;
		int i, j;
		ZeroMemory(TrackCount, TRACK_TYPES * sizeof(int));
		for (i = 0; i < InputCount; i++) {
			LPINPUTFILE InputFile;
			InputFile = GetInputFile(hWnd, i);
			for (j = 0; j < TRACK_TYPES; j++) {
				if (TrackCount[j] < InputFile->TypeCount[j]) {
					TrackCount[j] = InputFile->TypeCount[j];
				}
			}
		}
		CaretIndex = SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETCARETINDEX, 0, 0);
		ListIndex = 0;
		for (i = 0; i < TRACK_TYPES; i++) {
			for (j = 0; j < TrackCount[i]; j++) {
				LPARAM Data;
				Data = MAKELPARAM(j, i);
				while ((ListIndex < SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETCOUNT, 0, 0)) &&
				       (SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETITEMDATA, ListIndex, 0) < Data)) {
					SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_DELETESTRING, ListIndex, 0);
				}
				if ((ListIndex >= SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETCOUNT, 0, 0)) ||
				    (SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETITEMDATA, ListIndex, 0) > Data)) {
					TCHAR Mask[MAX_LOADSTRING];
					TCHAR Text[MAX_LOADSTRING + 16];
					LoadString(hInstance, IDS_VIDEOTRACK + i, Mask, MAX_LOADSTRING);
					wsprintf(Text, Mask, j + 1);
					SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_INSERTSTRING, ListIndex, Data);
					SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_SETSEL, TRUE, ListIndex);
				}
				if (SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETSEL, ListIndex, 0)) {
					HasTrackToExtract = TRUE;
				}
				ListIndex++;
			}
		}
		while (ListIndex < SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETCOUNT, 0, 0)) {
			SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_DELETESTRING, ListIndex, 0);
		}
		SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_SETCARETINDEX, CaretIndex, 0);
	} else {
		SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_RESETCONTENT, 0, 0);
	}
	EnableWindow(GetDlgItem(hWnd, IDC_EXTRACTTRACKS), HasTrackToExtract);
	UpdateWindow(hWnd);
}

BOOL AnalyzeAndAddFile(HWND hWnd, LPCTSTR FileName, BOOL ReportError)
{
	BOOL Unique = TRUE;
	int i;
	for (i = 0; i < GetInputCount(hWnd); i++) {
		if (! lstrcmpi(FileName, GetInputFile(hWnd, i)->FileName)) {
			Unique = FALSE;
		}
	}
	if (Unique) {
		LPINPUTFILE InputFile;
		InputFile = AnalyzeFile(FileName);
		if (InputFile) {
			AddInputFile(hWnd, InputFile);
			UpdateDialog(hWnd);
		} else if (ReportError) {
			TCHAR Mask[MAX_LOADSTRING];
			TCHAR Message[MAX_LOADSTRING + MAX_PATH];
			LoadString(hInstance, IDS_NOTANOGMKVFILE, Mask, MAX_LOADSTRING);
			wsprintf(Message, Mask, FileName);
			MessageBox(hWnd, Message, APPLICATION, MB_OK | MB_ICONEXCLAMATION);
		}
	}
}

#define MAX_OPEN_FILES 128

VOID DropDirectory(HWND hWnd, LPTSTR Path)
{
	TCHAR FindMask[MAX_PATH + 2];
	HANDLE FindHandle;
	WIN32_FIND_DATA FindData;
	wsprintf(FindMask, TEXT("%s\\*"), Path);
	FindHandle = FindFirstFile(FindMask, & FindData);
	while (FindHandle != INVALID_HANDLE_VALUE) {
		if (lstrcmp(FindData.cFileName, TEXT(".")) && lstrcmp(FindData.cFileName, TEXT("..")) &&
		    (lstrlen(Path) + lstrlen(FindData.cFileName) < MAX_PATH - 1)) {
			TCHAR FileName[MAX_PATH];
			wsprintf(FileName, TEXT("%s\\%s"), Path, FindData.cFileName);
			if (GetFileAttributes(FileName) & FILE_ATTRIBUTE_DIRECTORY) {
				DropDirectory(hWnd, FileName);
			} else {
				AnalyzeAndAddFile(hWnd, FileName, FALSE);
			}
		}
		if (! FindNextFile(FindHandle, & FindData)) {
			FindClose(FindHandle);
			FindHandle = INVALID_HANDLE_VALUE;
		}
	}
}

VOID DropFiles(HWND hWnd, HANDLE hDrop)
{
	UINT Count, i;
	TCHAR FileName[MAX_PATH];
	HCURSOR origCursor;
	SetForegroundWindow(hWnd);
	origCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	Count = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
	for (i = 0; i < Count; i++) {
		DragQueryFile(hDrop, i, FileName, MAX_PATH);
		if (GetFileAttributes(FileName) & FILE_ATTRIBUTE_DIRECTORY) {
			DropDirectory(hWnd, FileName);
		} else {
			AnalyzeAndAddFile(hWnd, FileName, TRUE);
		}
	}
	DragFinish(hDrop);
	SetCursor(origCursor);
}

VOID OpenFiles(HWND hWnd)
{
	TCHAR FileNames[MAX_PATH * MAX_OPEN_FILES] = TEXT("");
	OPENFILENAME ofn;
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize     = sizeof(OPENFILENAME);
	ofn.hwndOwner       = hWnd;
	ofn.hInstance       = hInstance;
	ofn.lpstrFilter     = TEXT("OGM and MKV files\0*.ogm;*.mkv\0\0");
	ofn.lpstrFile       = FileNames;
	ofn.nMaxFile        = MAX_PATH * MAX_OPEN_FILES;
	ofn.lpstrInitialDir = TEXT("");
	ofn.lpstrTitle      = TEXT("Open OGM/MKV files");
	ofn.Flags           = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_ALLOWMULTISELECT | OFN_EXPLORER;
	if (GetOpenFileName(& ofn)) {
		LPTSTR * FileList;
		int Count, i, j;
		HCURSOR origCursor;
		origCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
		FileList = GetOpenFileList(FileNames, & Count);
		for (i = 0; i < Count; i++) {
			AnalyzeAndAddFile(hWnd, FileList[i], TRUE);
		}
		FreeOpenFileList(FileList);
		SetCursor(origCursor);
	}
	EnableWindow(GetDlgItem(hWnd, IDC_EXTRACTTRACKS), SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETCOUNT, 0, 0));
}

VOID ExtractTracks(HWND hWnd, LPBOOL IsCancelled)
{
	TCHAR OutputDirectory[MAX_PATH];
	GetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, OutputDirectory, MAX_PATH);
	if (CreateDirectoryTree(OutputDirectory, FALSE, NULL)) {
		LPTSTR OrigDirectory;
		OrigDirectory = ChangeToDirectory(OutputDirectory);
		if (OrigDirectory) {
			BOOL TrackFirst; 
			BOOL NameTrackType;
			BOOL ConvertH264;
			TRACKMATRIX TrackMatrix;
			BOOL ok = TRUE;
			int i, j;
			TrackFirst    = (SendDlgItemMessage(hWnd, IDC_TRACKFIRST,    BM_GETCHECK, 0, 0) == BST_CHECKED);
			NameTrackType = (SendDlgItemMessage(hWnd, IDC_NAMETRACKTYPE, BM_GETCHECK, 0, 0) == BST_CHECKED);
			ConvertH264   = (SendDlgItemMessage(hWnd, IDC_CONVERTH264,   BM_GETCHECK, 0, 0) == BST_CHECKED);
			TrackMatrix.Count = 0;
			for (i = GetInputCount(hWnd) - 1; i >= 0; i--) {
				for (j = 0; j < TRACK_TYPES; j++) {
					if (TrackMatrix.Count < GetInputFile(hWnd, i)->TypeCount[j]) {
						TrackMatrix.Count = GetInputFile(hWnd, i)->TypeCount[j];
					}
				}
			}
			TrackMatrix.Type = (LPTRACKARRAY) FatalAlloc(TrackMatrix.Count * sizeof(TRACKARRAY));
			for (i = SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETCOUNT, 0, 0) - 1; i >= 0; i--) {
				if (SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETSEL, i, 0)) {
					LRESULT Data;
					Data = SendDlgItemMessage(hWnd, IDC_TRACKLIST, LB_GETITEMDATA, i, 0);
					TrackMatrix.Type[LOWORD(Data)][HIWORD(Data)] = TRUE;
				}
			}
			while (ok && GetInputCount(hWnd)) {
				LPINPUTFILE InputFile;
				InputFile = GetInputFile(hWnd, 0);
				for (i = GetInputCount(hWnd) - 1; i >= 0; i--) {
					SendDlgItemMessage(hWnd, IDC_FILELIST, LB_SETSEL, ! i, i);
				}
				UpdateDialog(hWnd);
				if (InputFile->IsOGM) {
					ok = ExtractOGMTracks(InputFile, TrackFirst, NameTrackType, ConvertH264, & TrackMatrix, IsCancelled);
				} else {
					ok = ExtractMKVTracks(InputFile, TrackFirst, NameTrackType, ConvertH264, & TrackMatrix, IsCancelled);
				}
				if (ok) {
					RemoveInputFile(hWnd, 0);
					UpdateDialog(hWnd);
				} else {
					TCHAR Message[MAX_LOADSTRING + MAX_PATH];
					if (! *IsCancelled) {
						TCHAR Mask[MAX_LOADSTRING];
						LoadString(hInstance, IDS_ERRORWHENEXTRACTING, Mask, MAX_LOADSTRING);
						wsprintf(Message, Mask, InputFile->FileName);
						MessageBox(hWnd, Message, APPLICATION, MB_OK | MB_ICONEXCLAMATION);
					}
				}
			}
			FatalFree(TrackMatrix.Type);
			SetCurrentDirectory(OrigDirectory);
			FatalFree(OrigDirectory);
		}
	} else {
		TCHAR Mask[MAX_LOADSTRING];
		TCHAR Message[MAX_LOADSTRING + MAX_PATH];
		LoadString(hInstance, IDS_CANNOTCREATEDIRECTORY, Mask, MAX_LOADSTRING);
		wsprintf(Message, Mask, OutputDirectory);
		MessageBox(hWnd, Message, APPLICATION, MB_OK | MB_ICONEXCLAMATION);
	}
}

VOID CancelExtracting(HWND hWnd, LPBOOL IsCancelled)
{
	TCHAR Message[MAX_LOADSTRING];
	LoadString(hInstance, IDS_EXTRACTINGCANCELLED, Message, MAX_LOADSTRING);
	if (MessageBox(hWnd, Message, APPLICATION, MB_YESNO | MB_ICONQUESTION) == IDYES) {
		*IsCancelled = TRUE;	
	}
}

VOID FindMKVToolnix(HWND hWnd)
{
	TCHAR path[MAX_PATH];
	TCHAR text[MAX_LOADSTRING];
	BOOL changed = FALSE;
	
	LoadString(hInstance, IDS_MKVTOOLNIXLOCATION, text, MAX_LOADSTRING);
	lstrcpy(path, Tools.MKVToolnix);
	if (BrowseForDirectory(hWnd, path, text, FALSE)) {
		changed = lstrcmpi(path, Tools.MKVToolnix);
		if (changed) {
			lstrcpy(Tools.MKVToolnix, path);
			lstrcpy(Tools.MKVInfo,    path);
			lstrcat(Tools.MKVInfo,    TEXT("\\MKVInfo.exe"));
			lstrcpy(Tools.MKVExtract, path);
			lstrcat(Tools.MKVExtract, TEXT("\\MKVExtract.exe"));
		}
	}
	
	if (changed) {
		char filename[MAX_PATH];
		FILE * f;
		if (GetFullPathInSpecialFolder(CSIDL_APPDATA, TEXT("OGMcleaver"), NULL, TEXT("MKVToolnix"), path, MAX_PATH, TRUE, NULL)) {
			#ifdef UNICODE
			WideCharToMultiByte(CP_ACP, 0, path, -1, filename, MAX_PATH, NULL, NULL);
			#else
			lstrcpy(filename, path);
			#endif
			f = fopen(filename, "wt");
			if (f) {
				#ifdef UNICODE
				WideCharToMultiByte(CP_ACP, 0, Tools.MKVToolnix, -1, filename, MAX_PATH, NULL, NULL);
				#else
				lstrcpy(filename, Tools.MKVToolnix);
				#endif
				fprintf(f, filename);
				fclose(f);
			}
		}
	}
}

VOID AboutBox(HWND hWnd)
{
	MSGBOXPARAMS mbp;
	TCHAR Text[MAX_LOADSTRING];
	TCHAR AboutText[MAX_LOADSTRING * 2] = APP_VERSION TEXT("\n");
	
	LoadString(hInstance, IDS_ABOUT, Text, MAX_LOADSTRING);
	lstrcat(AboutText, Text);
	
	ZeroMemory(& mbp, sizeof(MSGBOXPARAMS));
	mbp.cbSize      = sizeof(MSGBOXPARAMS);
	mbp.hwndOwner   = hWnd;
	mbp.hInstance   = hInstance;
	mbp.lpszText    = AboutText;
	mbp.lpszCaption = APPLICATION;
	mbp.dwStyle     = MB_USERICON | MB_OK;
	mbp.lpszIcon    = MAKEINTRESOURCE(IDI_ICON);
	MessageBoxIndirect(& mbp);
}

VOID MeasureItem(HWND hWnd, LPMEASUREITEMSTRUCT mis)
{
	HDC DC; 
	HFONT Font;
	HGDIOBJ OrigFont;
	DC = GetDC(hWnd);
	Font = (HFONT) SendMessage(hWnd, WM_GETFONT, 0, 0);
	OrigFont = SelectObject(DC, Font);
	mis->itemWidth  = 0;
	mis->itemHeight = 0;
	
	if (mis->CtlID == IDC_FILELIST) {
		LPINPUTFILE InputFile = (LPINPUTFILE) mis->itemData;
		if (InputFile) {
			TCHAR Mask[MAX_LOADSTRING];
			TCHAR Text[MAX_LOADSTRING + 48];
			SIZE Size;
			LoadString(hInstance, IDS_VIDEOAUDIOSUBTITLETRACKS, Mask, MAX_LOADSTRING);
			wsprintf(Text, Mask, InputFile->TypeCount[VIDEO_TRACK], InputFile->TypeCount[AUDIO_TRACK], InputFile->TypeCount[SUBTITLE_TRACK]);
			GetTextExtentPoint32(DC, InputFile->FileName, lstrlen(InputFile->FileName), & Size);
			mis->itemWidth  = Size.cx;
			mis->itemHeight = Size.cy;
			GetTextExtentPoint32(DC, Text, lstrlen(Text), & Size);
			if (mis->itemWidth < Size.cx) mis->itemWidth = Size.cx;
			mis->itemHeight += Size.cy;
		}
		mis->itemWidth  += 4;
		mis->itemHeight += 5;

	} else if (mis->CtlID == IDC_TRACKLIST) {
		int TrackType;
		int TrackNo;
		TCHAR Mask[MAX_LOADSTRING];
		TCHAR Text[MAX_LOADSTRING + 12];
		SIZE Size;
		TrackType = HIWORD(mis->itemData);
		TrackNo   = LOWORD(mis->itemData) + 1;
		LoadString(hInstance, IDS_VIDEOTRACK + TrackType, Mask, MAX_LOADSTRING);
		wsprintf(Text, Mask, TrackNo);
		GetTextExtentPoint32(DC, Text, lstrlen(Text), & Size);
		mis->itemWidth  = Size.cx + 4;
		mis->itemHeight = Size.cy + 2;
		if (! mis->itemID) mis->itemHeight++;
	}
		
	SelectObject(DC, OrigFont);
	ReleaseDC(hWnd, DC);
}

VOID DrawItem(HWND hWnd, LPDRAWITEMSTRUCT dis)
{
	RECT ItemRect;
	HDC DC;
	HBITMAP Bitmap;
	COLORREF BgColor;
	COLORREF FgColor;
	HFONT Font;
	HBRUSH Brush;
	
	ItemRect.left   = 0;
	ItemRect.top    = 0;
	ItemRect.right  = dis->rcItem.right - dis->rcItem.left;
	ItemRect.bottom = dis->rcItem.bottom - dis->rcItem.top;
	
	DC = CreateCompatibleDC(dis->hDC);
	Bitmap = CreateCompatibleBitmap(dis->hDC, ItemRect.right, ItemRect.bottom);
	SelectObject(DC, Bitmap);

	if ((dis->CtlID == IDC_FILELIST) && (dis->itemState & ODS_SELECTED)) {
		BgColor = GetSysColor(COLOR_HIGHLIGHT);
		FgColor = GetSysColor(COLOR_HIGHLIGHTTEXT);
	} else {
		BgColor = GetSysColor(COLOR_WINDOW);
		FgColor = GetSysColor(COLOR_WINDOWTEXT);
	}
	Font = (HFONT) SendMessage(hWnd, WM_GETFONT, 0, 0);
	Brush = CreateSolidBrush(BgColor);
	FillRect(DC, & ItemRect, Brush);
	DeleteObject(Brush);
	
	if (dis->CtlID == IDC_FILELIST) {
		LPINPUTFILE InputFile = (LPINPUTFILE) dis->itemData;
		if (InputFile) {
			LOGFONT LogFont;
			TCHAR Mask[MAX_LOADSTRING];
			TCHAR Text[MAX_LOADSTRING + 48];
			SIZE Size;
			RECT Rect;
			HFONT TextFont;
			HPEN Pen;
			GetObject(Font, sizeof(LOGFONT), & LogFont);
			LogFont.lfWeight = FW_BOLD;
			TextFont = CreateFontIndirect(& LogFont);
			SelectObject(DC, TextFont);
			LoadString(hInstance, IDS_VIDEOAUDIOSUBTITLETRACKS, Mask, MAX_LOADSTRING);
			wsprintf(Text, Mask, InputFile->TypeCount[VIDEO_TRACK], InputFile->TypeCount[AUDIO_TRACK], InputFile->TypeCount[SUBTITLE_TRACK]);
			GetTextExtentPoint32(DC, InputFile->FileName, lstrlen(InputFile->FileName), & Size);
			Rect.left   = ItemRect.left  + 2;
			Rect.top    = ItemRect.top   + 2;
			Rect.right  = ItemRect.right - 2;
			Rect.bottom = Rect.top + Size.cy;
			SetTextColor(DC, FgColor);
			SetBkColor(DC, BgColor);
			DrawText(DC, InputFile->FileName, lstrlen(InputFile->FileName), 
				& Rect, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_NOPREFIX | DT_PATH_ELLIPSIS);
			DeleteObject(TextFont);
			LogFont.lfWeight = FW_NORMAL;
			TextFont = CreateFontIndirect(& LogFont);
			SelectObject(DC, TextFont);
			GetTextExtentPoint32(DC, Text, lstrlen(Text), & Size);
			Rect.top    = Rect.bottom;
			Rect.bottom = Rect.top + Size.cy;
			DrawText(DC, Text, lstrlen(Text), & Rect, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_NOPREFIX);
			DeleteObject(TextFont);
			Pen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_GRAYTEXT));
			SelectObject(DC, Pen);
			MoveToEx(DC, ItemRect.left, ItemRect.bottom - 1, NULL);
			LineTo(DC, ItemRect.right, ItemRect.bottom - 1);
			DeleteObject(Pen);
		}
		ItemRect.bottom--;
	
	} else if (dis->CtlID == IDC_TRACKLIST) {
		int TrackType;
		int TrackNo;
		RECT Rect;
		TCHAR Mask[MAX_LOADSTRING];
		TCHAR Text[MAX_LOADSTRING + 12];
		if (! dis->itemID) ItemRect.top++;
		TrackType = HIWORD(dis->itemData);
		TrackNo   = LOWORD(dis->itemData) + 1;
		Rect.left   = ItemRect.left   + 2;
		Rect.top    = ItemRect.top    + 1;
		Rect.bottom = ItemRect.bottom - 1;
		Rect.right  = Rect.left + (Rect.bottom - Rect.top);
		DrawFrameControl(DC, & Rect, DFC_BUTTON, DFCS_BUTTONCHECK	| ((dis->itemState & ODS_SELECTED) ? DFCS_CHECKED : 0));
		LoadString(hInstance, IDS_VIDEOTRACK + TrackType, Mask, MAX_LOADSTRING);
		wsprintf(Text, Mask, TrackNo);
		Rect.left = Rect.right + 2;
		Rect.right = ItemRect.right - 2;
		SelectObject(DC, Font);
		DrawText(DC, Text, lstrlen(Text), & Rect, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_NOPREFIX);
	}
	
	if (dis->itemState & ODS_FOCUS) {
		DrawFocusRect(DC, & ItemRect);
	}
		
	BitBlt(dis->hDC, dis->rcItem.left, dis->rcItem.top, dis->rcItem.right - dis->rcItem.left, dis->rcItem.bottom - dis->rcItem.top, DC, 0, 0, SRCCOPY);
	DeleteObject(Bitmap);
	DeleteObject(DC);
}

VOID InitDialog(HWND hWnd)
{
	BOOL TrackFirst    = TRUE;
	BOOL NameTrackType = FALSE;
	BOOL ConvertH264   = TRUE;
	WINDOWPLACEMENT Place;
	LONG X, Y;
	LPTSTR CfgFileName;

	SetWindowText(hWnd, APP_VERSION);
	SetClassLong(hWnd, GCL_HICON, (LONG) LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON)));
	GetWindowPlacement(hWnd, & Place);
	X = Place.rcNormalPosition.left;
	Y = Place.rcNormalPosition.top;

	CfgFileName = GetFullPathInSpecialFolder(CSIDL_APPDATA, TEXT("OGMcleaver"), NULL, TEXT("config"), NULL, 0, FALSE, NULL);
	if (CfgFileName) {
		HANDLE CfgFileHandle;
		CfgFileHandle = CreateFile(CfgFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (CfgFileHandle != INVALID_HANDLE_VALUE) {
			DWORD FileSize;
			FileSize = GetFileSize(CfgFileHandle, NULL);
			if (FileSize != 0xFFFFFFFF) {
				char * config;
				DWORD BytesRead;
				config = (char *) FatalAlloc(FileSize + 1);
				if (ReadFile(CfgFileHandle, config, FileSize, & BytesRead, NULL) && (BytesRead == FileSize)) {
					sscanf(config, "%d,%d,%d,%d,%d", & TrackFirst, & NameTrackType, & ConvertH264, & X, & Y);
				}
			}
			CloseHandle(CfgFileHandle);
		}
		FatalFree(CfgFileName);
	}
		
	SendDlgItemMessage(hWnd, IDC_TRACKFIRST,    BM_SETCHECK, (TrackFirst    ? BST_CHECKED : BST_UNCHECKED), 0);
	SendDlgItemMessage(hWnd, IDC_TRACKLAST,     BM_SETCHECK, (TrackFirst    ? BST_UNCHECKED : BST_CHECKED), 0);
	SendDlgItemMessage(hWnd, IDC_NAMETRACKTYPE, BM_SETCHECK, (NameTrackType ? BST_CHECKED : BST_UNCHECKED), 0);
	SendDlgItemMessage(hWnd, IDC_CONVERTH264,   BM_SETCHECK, (ConvertH264   ? BST_CHECKED : BST_UNCHECKED), 0);
	Place.rcNormalPosition.right -= (Place.rcNormalPosition.left - X);
	Place.rcNormalPosition.bottom -= (Place.rcNormalPosition.top - Y);
	Place.rcNormalPosition.left = X;
	Place.rcNormalPosition.top = Y;
	SetWindowPlacement(hWnd, & Place);
	UpdateDialog(hWnd);
}

VOID ExitDialog(HWND hWnd)
{
	BOOL TrackFirst;
	BOOL NameTrackType;
	BOOL ConvertH264;
	WINDOWPLACEMENT Place;
	TCHAR Config[256];
	char config[256];
	LPTSTR CfgFileName;
	
	TrackFirst    = (SendDlgItemMessage(hWnd, IDC_TRACKFIRST,    BM_GETCHECK, 0, 0) == BST_CHECKED);
	NameTrackType = (SendDlgItemMessage(hWnd, IDC_NAMETRACKTYPE, BM_GETCHECK, 0, 0) == BST_CHECKED);
	ConvertH264   = (SendDlgItemMessage(hWnd, IDC_CONVERTH264,   BM_GETCHECK, 0, 0) == BST_CHECKED);
	GetWindowPlacement(hWnd, & Place);
	wsprintf(Config, "%d,%d,%d,%d,%d", TrackFirst, NameTrackType, ConvertH264, Place.rcNormalPosition.left, Place.rcNormalPosition.top);
	#ifdef UNICODE
	WideCharToMultiByte(CP_ACP, 0, Config, -1, config, 256, NULL, NULL);
	#else
	lstrcpy(config, Config);
	#endif
	
	CfgFileName = GetFullPathInSpecialFolder(CSIDL_APPDATA, TEXT("OGMcleaver"), NULL, TEXT("config"), NULL, 0, TRUE, NULL);
	if (CfgFileName) {
		HANDLE CfgFileHandle;
		CfgFileHandle = CreateFile(CfgFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (CfgFileHandle != INVALID_HANDLE_VALUE) {
			DWORD BytesWritten;
			WriteFile(CfgFileHandle, config, lstrlen(Config), & BytesWritten, NULL);
			CloseHandle(CfgFileHandle);
		}
		FatalFree(CfgFileName);
	}
}

BOOL CALLBACK MainDialogProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static BOOL IsCancelled;
	static BOOL IsExtracting;
	
	switch (message)
	{
	case WM_INITDIALOG:
		IsCancelled  = FALSE;
		IsExtracting = FALSE;
		InitDialog(hWnd);
		DragAcceptFiles(hWnd, TRUE);
		return TRUE;

	case WM_DROPFILES:
		DropFiles(hWnd, (HANDLE) wParam);
		return TRUE;
		
	case WM_COMMAND:
		switch (LOWORD(wParam)) 
		{
		case IDC_OPENFILES:
			if (HIWORD(wParam) == BN_CLICKED) {
				OpenFiles(hWnd);
			}
			return TRUE;

		case IDC_REMOVEFILES:
			if (HIWORD(wParam) == BN_CLICKED) {
				RemoveSelectedInputFiles(hWnd);
				UpdateDialog(hWnd);
			}
			return TRUE;
			
		case IDC_FINDMKVTOOLNIX:
			if (HIWORD(wParam) == BN_CLICKED) {
				FindMKVToolnix(hWnd);
			}
			return TRUE;
			
		case IDC_ABOUT:
			if (HIWORD(wParam) == BN_CLICKED) {
				AboutBox(hWnd);
			}
			return TRUE;

		case IDC_OUTPUTDIRECTORY:
			if (HIWORD(wParam) == EN_KILLFOCUS) {
				TCHAR Directory[MAX_PATH], FullPath[MAX_PATH];
				GetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, Directory, MAX_PATH);
				if (GetFullPathName(Directory, MAX_PATH, FullPath, NULL)) {
					SetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, FullPath);
				} else {
					SetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, TEXT(""));
				}
			}
			return TRUE;

		case IDC_BROWSEOUTPUTDIRECTORY:
			if (HIWORD(wParam) == BN_CLICKED) {
				TCHAR Directory[MAX_PATH];
				TCHAR Caption[MAX_LOADSTRING];
				GetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, Directory, MAX_PATH);
				LoadString(hInstance, IDS_BROWSEOUTPUTDIRECTORY, Caption, MAX_LOADSTRING);
				if (BrowseForDirectory(hWnd, Directory, Caption, TRUE)) {
					SetDlgItemText(hWnd, IDC_OUTPUTDIRECTORY, Directory);
				}
			}
			return TRUE;

		case IDC_EXTRACTTRACKS:
			if (HIWORD(wParam) == BN_CLICKED) {
				if (IsExtracting) {
					CancelExtracting(hWnd, & IsCancelled);
				} else {
					TCHAR OrigText[MAX_LOADSTRING];
					TCHAR NewText[MAX_LOADSTRING];
					
					IsExtracting = TRUE;
					
					EnableMenuItem(GetSystemMenu(hWnd, FALSE), SC_CLOSE, MF_BYCOMMAND | MF_DISABLED);
					EnableWindow(GetDlgItem(hWnd, IDC_FILEGROUP),             FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_OPENFILES),             FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_REMOVEFILES),           FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_FILELIST),              FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_OUTPUTGROUP),           FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_OUTPUTDIRECTORY),       FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_BROWSEOUTPUTDIRECTORY), FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKGROUP),            FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKLIST),             FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_FILENAMEGROUP),         FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKFIRST),            FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKLAST),             FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_NAMETRACKTYPE),         FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_OPTIONGROUP),           FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_CONVERTH264),           FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_FINDMKVTOOLNIX),        FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_ABOUT),                 FALSE);
					
					GetDlgItemText(hWnd, IDC_EXTRACTTRACKS, OrigText, MAX_LOADSTRING);
					LoadString(hInstance, IDS_CANCEL, NewText, MAX_LOADSTRING);
					SetDlgItemText(hWnd, IDC_EXTRACTTRACKS, NewText);
					LoadString(hInstance, IDS_EXTRACTINGTRACKS, NewText, MAX_LOADSTRING);
					SetWindowText(hWnd, NewText);
					ExtractTracks(hWnd, & IsCancelled);
					SetWindowText(hWnd, APP_VERSION);
					SetDlgItemText(hWnd, IDC_EXTRACTTRACKS, OrigText);
					
					EnableMenuItem(GetSystemMenu(hWnd, FALSE), SC_CLOSE, MF_BYCOMMAND | MF_ENABLED);
					EnableWindow(GetDlgItem(hWnd, IDC_FILEGROUP),             TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_OPENFILES),             TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_REMOVEFILES),           SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETSELCOUNT, 0, 0));
					EnableWindow(GetDlgItem(hWnd, IDC_FILELIST),              TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_OUTPUTGROUP),           TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_OUTPUTDIRECTORY),       TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_BROWSEOUTPUTDIRECTORY), TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKGROUP),            TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKLIST),             TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_FILENAMEGROUP),         TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKFIRST),            TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_TRACKLAST),             TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_NAMETRACKTYPE),         TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_OPTIONGROUP),           TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_CONVERTH264),           TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_FINDMKVTOOLNIX),        TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_ABOUT),                 TRUE);

					IsCancelled  = FALSE;
					IsExtracting = FALSE;
				}
			}
			return TRUE;
		
		case IDC_FILELIST:
			if (HIWORD(wParam) == LBN_SELCHANGE) {
				EnableWindow(GetDlgItem(hWnd, IDC_REMOVEFILES), SendDlgItemMessage(hWnd, IDC_FILELIST, LB_GETSELCOUNT, 0, 0));
			}
			return TRUE;

		case IDC_TRACKLIST:
			if (HIWORD(wParam) == LBN_SELCHANGE) {
				UpdateDialog(hWnd);
			}
			return TRUE;
		}
		return FALSE;
		
	case WM_MEASUREITEM:
		MeasureItem(hWnd, (LPMEASUREITEMSTRUCT) lParam);
		return TRUE;

	case WM_DRAWITEM:
		DrawItem(hWnd, (LPDRAWITEMSTRUCT) lParam);
		return TRUE;

	case WM_QUERYENDSESSION:
		if (IsExtracting) {
			CancelExtracting(hWnd, & IsCancelled);
			return IsCancelled;
		}
		return TRUE;
		
	case WM_CLOSE:
		if (IsExtracting) return TRUE;
	case WM_ENDSESSION:
		ExitDialog(hWnd);
		EndDialog(hWnd, 0);
		return TRUE;
		
	}
	return FALSE;
}

int WINAPI WinMain(HINSTANCE Instance, HINSTANCE PrevInstance, LPSTR CmdLine, int CmdShow)
{
	hInstance = Instance;

	InitializeTools();
	CoInitialize(NULL);

	DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_EXTRACTDIALOG), HWND_DESKTOP, 
		(DLGPROC) MainDialogProc, (LPARAM) & CmdShow);
		
	CoUninitialize();

	return 0;
}
