#include "winutils.h"

#include <shlobj.h>
#include <stdio.h>

LPVOID FatalAlloc(DWORD Size)
{
	LPVOID Ptr;
	Ptr = (LPVOID) calloc(1, Size); // LocalAlloc(LMEM_FIXED | LMEM_ZEROINIT, Size);
	if (! Ptr) ExitProcess(1);
	return Ptr;
}

LPVOID FatalReAlloc(LPVOID Ptr, DWORD Size)
{
	if (! Ptr) return FatalAlloc(Size);
	Ptr = (LPVOID) realloc(Ptr, Size); // LocalReAlloc((HLOCAL) Ptr, Size, LMEM_ZEROINIT);
	if (! Ptr) ExitProcess(1);
	return Ptr;
}

VOID FatalFree(LPVOID Ptr)
{
	if (Ptr) free(Ptr);
}

LPTSTR CopyStr(LPTSTR Dst, LPCTSTR Src)
{
	if (Src) {
		Dst = FatalReAlloc(Dst, (lstrlen(Src) + 1) * sizeof(TCHAR));
		lstrcpy(Dst, Src);
		return Dst;
	} else {
		FatalFree(Dst);
		return NULL;
	}
}

LPTSTR ConcatStr(LPTSTR Dst, LPCTSTR Src)
{
	int offset;
	offset = (Dst ? lstrlen(Dst) : 0);
	Dst = FatalReAlloc(Dst, (offset + lstrlen(Src) + 1) * sizeof(TCHAR));
	lstrcpy(Dst + offset, Src);
	return Dst;
}

LPTSTR DoubleStr(double d)
{
	char result[32];
	size_t size;
	LPTSTR Result;
	sprintf(result, "%f", d);
	size = strlen(result) + 1;
	Result = FatalAlloc(size * sizeof(TCHAR));
#ifdef UNICODE
	MultiByteToWideChar(CP_ACP, 0, result, -1, Result, size);
#else
	lstrcpy(Result, result);
#endif
	return Result;
}

LPTSTR GetAllocFullPathName(LPCTSTR lpFileName,	LPTSTR * lpFilePart)
{
	LPTSTR fullPathName;
	DWORD size;
	
	size = GetFullPathName(lpFileName, 0, NULL, NULL);
	if (! size) return NULL;

	fullPathName = (LPTSTR) FatalAlloc(size * sizeof(TCHAR));

	if (! GetFullPathName(lpFileName, size, fullPathName, lpFilePart)) {
		FatalFree(fullPathName);
		return NULL;
	}

	return fullPathName;
}

BOOL CreateDirectoryTree(LPCTSTR lpFullPathName, 
                         BOOL bLastTokenIsFileName, 
                         LPSECURITY_ATTRIBUTES lpSecurityAttributes)
{
	LPTSTR path;
	path = GetAllocFullPathName(lpFullPathName, NULL);
	if (path) {
		BOOL ok = TRUE;
		int length, i;
		TCHAR last = TEXT(':');
		length = lstrlen(path);
		for (i = 1; ok && i <= length; i++) {
			if ((i < length && last != TEXT(':') && path[i] == TEXT('\\'))
			 || (i == length && !bLastTokenIsFileName)) {
				DWORD fileAttributes;
				if (i < length) path[i] = 0;
				fileAttributes = GetFileAttributes(path);
				if (fileAttributes == 0xFFFFFFFF || ! (fileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
					ok = CreateDirectory(path, lpSecurityAttributes);
				}
				if (i < length) path[i] = TEXT('\\');
			}
			if (i < length) last = path[i];
		}
		FatalFree(path);
		return ok;
	}
	return FALSE;
}

int CALLBACK BrowseForDirectoryCallbackProc(HWND hWnd, UINT message, LPARAM lParam, LPARAM lpData)
{
	switch (message) 
	{
	case BFFM_INITIALIZED:
		SendMessage(hWnd, BFFM_SETSELECTION, TRUE, lpData);
		break;
	}
	return 0;
}

BOOL BrowseForDirectory(HWND hWnd, LPTSTR Folder, LPCTSTR Title, BOOL CreateButton)
{
	BROWSEINFO BrowseInfo;
	LPITEMIDLIST IdList;
	BOOL ok;

	BrowseInfo.hwndOwner      = hWnd;
	BrowseInfo.pidlRoot       = NULL;
	BrowseInfo.pszDisplayName = NULL;
	BrowseInfo.lpszTitle      = Title;
	BrowseInfo.ulFlags        = BIF_RETURNONLYFSDIRS | (CreateButton ? BIF_NEWDIALOGSTYLE : 0);
	BrowseInfo.lpfn           = (BFFCALLBACK) BrowseForDirectoryCallbackProc;
	BrowseInfo.lParam         = (LPARAM) Folder;
	
	IdList = SHBrowseForFolder(& BrowseInfo);
	ok = (IdList && SHGetPathFromIDList(IdList, Folder));
	if (IdList) CoTaskMemFree(IdList);
	return ok;
}

LPTSTR ChangeToDirectory(LPCTSTR Directory)
{
	DWORD  Size;
	LPTSTR CurrentDirectory;
	
	Size = GetCurrentDirectory(0, NULL);
	if (! Size) return NULL;
	CurrentDirectory = FatalAlloc(Size * sizeof(TCHAR));
	if (GetCurrentDirectory(Size, CurrentDirectory)) {
		if (SetCurrentDirectory(Directory)) {
			return CurrentDirectory;
		}
	}
	FatalFree(CurrentDirectory);
	return NULL;
}

BOOL ChangeDirectoryBack(LPTSTR OrigDirectory)
{
	BOOL Result = SetCurrentDirectory(OrigDirectory);
	FatalFree(OrigDirectory);
	return Result;
}

BOOL CleanDirectory(LPCTSTR Directory)
{
	LPTSTR OrigDirectory;
	HANDLE FindHandle;
	WIN32_FIND_DATA FindData;
	BOOL ok = TRUE;

	OrigDirectory = ChangeToDirectory(Directory);
	if (! OrigDirectory) return FALSE;
	FindHandle = FindFirstFile(TEXT("*"), & FindData);
	ok = (FindHandle != INVALID_HANDLE_VALUE);
	if (ok) do {
		if (lstrcmp(FindData.cFileName, TEXT(".")) && lstrcmp(FindData.cFileName, TEXT(".."))) {
			if ((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
				LPTSTR SubDirectory;
				SubDirectory = FatalAlloc((lstrlen(Directory) + lstrlen(FindData.cFileName) + 2) * sizeof(TCHAR));
				wsprintf(SubDirectory, TEXT("%s\\%s"), Directory, FindData.cFileName);
				ok = CleanDirectory(SubDirectory) 
				  && SetFileAttributes(FindData.cFileName, FILE_ATTRIBUTE_NORMAL) 
				  && RemoveDirectory(FindData.cFileName);
				FatalFree(SubDirectory);
			} else {
				ok = SetFileAttributes(FindData.cFileName, FILE_ATTRIBUTE_NORMAL)
				  && DeleteFile(FindData.cFileName);
			}
		}
	} while (ok && FindNextFile(FindHandle, & FindData));
	if (FindHandle != INVALID_HANDLE_VALUE) FindClose(FindHandle);
	ChangeDirectoryBack(OrigDirectory);
	return ok;
}

BOOL DeleteFilesFromDirectory(LPCTSTR Directory, LPCTSTR FindMask)
{
	LPTSTR OrigDirectory;
	HANDLE FindHandle;
	WIN32_FIND_DATA FindData;
	BOOL ok = TRUE;

	OrigDirectory = ChangeToDirectory(Directory);
	if (! OrigDirectory) return FALSE;
	FindHandle = FindFirstFile(FindMask, & FindData);
	ok = (FindHandle != INVALID_HANDLE_VALUE);
	if (ok) do {
		if (lstrcmp(FindData.cFileName, TEXT(".")) && lstrcmp(FindData.cFileName, TEXT(".."))
			&& ((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)) {
			ok = SetFileAttributes(FindData.cFileName, FILE_ATTRIBUTE_NORMAL)
			  && DeleteFile(FindData.cFileName);
		} 
	} while (ok && FindNextFile(FindHandle, & FindData));
	if (FindHandle != INVALID_HANDLE_VALUE) FindClose(FindHandle);
	SetCurrentDirectory(OrigDirectory);
	FatalFree(OrigDirectory);
	return ok;
}

LPTSTR GetFullPathInSpecialFolder(int csidl, LPCTSTR lpMainDirectory, LPCTSTR lpSubDirectory, LPCTSTR lpFileName, 
            LPTSTR lpFullPathName, UINT uSize, BOOL bCreateDirectory, LPSECURITY_ATTRIBUTES lpSecurityAttributes)
{
	LPITEMIDLIST idList;
	TCHAR rootDirectory[MAX_PATH];
	BOOL ok;
	
	ok = (SHGetSpecialFolderLocation(NULL, csidl, &idList) == NOERROR);
	if (ok) {
		ok = SHGetPathFromIDList(idList, rootDirectory);
		CoTaskMemFree(idList);
	}
	if (!ok) {
		lpMainDirectory = 0;
		ok = (GetModuleFileName(NULL, rootDirectory, MAX_PATH) > 0);
		if (ok) {
			int i;
			for (i = lstrlen(rootDirectory) - 1; i >= 0 && rootDirectory[i] != TEXT('\\'); i--);
			if (i > 0) rootDirectory[i] = 0;
		}
	}
	
	if (ok) {
		UINT length;
		LPTSTR pathName;
		LPTSTR fullPathName;
		length = lstrlen(rootDirectory)
		       + (lpMainDirectory && *lpMainDirectory ? lstrlen(lpMainDirectory) + 1 : 0)
		       + (lpSubDirectory && *lpSubDirectory ? lstrlen(lpSubDirectory) + 1 : 0)
		       + (lpFileName && *lpFileName ? lstrlen(lpFileName) + 1 : 0)
		       + 1;
		pathName = (LPTSTR) FatalAlloc(length * sizeof(TCHAR));
		lstrcpy(pathName, rootDirectory);
		if (lpMainDirectory && *lpMainDirectory) {
			lstrcat(pathName, TEXT("\\"));
			lstrcat(pathName, lpMainDirectory);
		}
		if (lpSubDirectory && *lpSubDirectory) {
			lstrcat(pathName, TEXT("\\"));
			lstrcat(pathName, lpSubDirectory);
		}
		if (lpFileName && *lpFileName) {
			lstrcat(pathName, TEXT("\\"));
			lstrcat(pathName, lpFileName);
		}
		fullPathName = GetAllocFullPathName(pathName, NULL);
		FatalFree(pathName);
		ok = (fullPathName != NULL);

		if (ok) {
			length = lstrlen(fullPathName);
			if (lpFullPathName) {
				ok = (length < uSize);
			}
			if (ok && bCreateDirectory) {
				ok = CreateDirectoryTree(fullPathName, (lpFileName != NULL), lpSecurityAttributes);
			}
			if (lpFullPathName) {
				if (ok) lstrcpy(lpFullPathName, fullPathName);
				FatalFree(fullPathName);
			} else if (ok) {
				lpFullPathName = fullPathName;
			} else {
				FatalFree(fullPathName);
			}
		}
	}
	
	if (ok) {
		return lpFullPathName;
	} else {
		return NULL;
	}
}

LPTSTR * GetOpenFileList(LPTSTR FileNames, int * FileCount)
{
	LPTSTR * List = NULL;
	int Size  = 0;
	int Count = 0;
	int pos;
	pos = lstrlen(FileNames);
	if (FileNames[pos + 1]) {
		do {
			if (Count >= Size - 2) {
				Size = (Size ? Size * 2 : 8);
				List = FatalReAlloc(List, Size * sizeof(LPTSTR));
			}
			pos++;
			List[Count] = FatalAlloc((lstrlen(FileNames) + lstrlen(FileNames + pos) + 2) * sizeof(TCHAR));
			wsprintf(List[Count], TEXT("%s\\%s"), FileNames, FileNames + pos);
			pos += lstrlen(FileNames + pos);
			Count++;
		} while (FileNames[pos + 1]);
		List[Count] = NULL;
	} else {
		List = FatalReAlloc(List, 2 * sizeof(LPTSTR));
		List[0] = (LPTSTR) FatalAlloc((lstrlen(FileNames) + 1) * sizeof(TCHAR));
		lstrcpy(List[0], FileNames);
		Count = 1;
	}
	if (FileCount) *FileCount = Count;
	return List;
}

VOID FreeOpenFileList(LPTSTR * FileList)
{
	int i = 0;
	while (FileList[i]) FatalFree(FileList[i++]);
	FatalFree(FileList);
}

VOID ExecuteMessageLoop()
{
	MSG msg;
	while (PeekMessage(& msg, NULL, 0, 0, PM_REMOVE)) {
		TranslateMessage(& msg); 
		DispatchMessage(& msg); 
	}
}

BOOL ExecuteCommand(LPCTSTR ApplicationName, LPCTSTR CommandLine, 
                    LPCTSTR InFileName, LPCTSTR OutFileName, LPCTSTR ErrFileName,
                    LPBOOL IsCancelled, LPBOOL WaitingForUser)
{
	BOOL FreeCommandLine = FALSE;
	LPTSTR FullCommandLine = NULL;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	HANDLE InFile = NULL;
	HANDLE OutFile = NULL;
	HANDLE ErrFile = NULL;
	SECURITY_ATTRIBUTES sa;
	DWORD ExitCode = 0;

	ZeroMemory(& sa, sizeof(SECURITY_ATTRIBUTES));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.bInheritHandle = TRUE;

	if (ApplicationName && CommandLine) {
		FullCommandLine = (LPTSTR) FatalAlloc((lstrlen(ApplicationName) + 4 + lstrlen(CommandLine)) * sizeof(TCHAR));
		if (! FullCommandLine) return FALSE;
		wsprintf(FullCommandLine, TEXT("\"%s\" %s"), ApplicationName, CommandLine);
		FreeCommandLine = TRUE;
	} else if (ApplicationName) {
		FullCommandLine = (LPTSTR) ApplicationName;
	} else {
		FullCommandLine = (LPTSTR) CommandLine;
	}

	if (InFileName) {
		InFile = CreateFile(InFileName, GENERIC_READ, FILE_SHARE_READ, 
			& sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (InFile == INVALID_HANDLE_VALUE) {
			if (FreeCommandLine) FatalFree(FullCommandLine);
			return FALSE;
		}
	}
	if (OutFileName) {
		OutFile = CreateFile(OutFileName, GENERIC_WRITE, FILE_SHARE_READ, 
			& sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (OutFile == INVALID_HANDLE_VALUE) {
			if (FreeCommandLine) FatalFree(FullCommandLine);
			if (InFile) CloseHandle(InFile);
			return FALSE;
		}
	}
	if (ErrFileName) {
		ErrFile = CreateFile(ErrFileName, GENERIC_WRITE, FILE_SHARE_READ, 
			& sa, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (ErrFile == INVALID_HANDLE_VALUE) {
			if (FreeCommandLine) FatalFree(FullCommandLine);
			if (InFile) CloseHandle(InFile);
			if (OutFile) CloseHandle(OutFile);
			return FALSE;
		} else {
			SetFilePointer(ErrFile, 0, NULL, FILE_END);
		}
	}

	ZeroMemory(& si, sizeof(STARTUPINFO));
	si.cb          = sizeof(STARTUPINFO);
	si.dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
	si.wShowWindow = SW_SHOWNORMAL/*SW_SHOWNOACTIVATE*/;
	si.hStdInput   = (InFile  ? InFile  : GetStdHandle(STD_INPUT_HANDLE));
	si.hStdOutput  = (OutFile ? OutFile : GetStdHandle(STD_OUTPUT_HANDLE));
	si.hStdError   = (ErrFile ? ErrFile : GetStdHandle(STD_ERROR_HANDLE));

	if (CreateProcess(ApplicationName, FullCommandLine, NULL, NULL, TRUE, 
		              CREATE_NO_WINDOW, NULL, NULL, &si, &pi)) {
		while (WaitForSingleObject(pi.hProcess, 10) == WAIT_TIMEOUT) {
			if (IsCancelled && *IsCancelled) {
				TerminateProcess(pi.hProcess, 0xFFFFFFFF);
				ExitCode = 0xFFFFFFFF;
			} else {
				if (IsCancelled) ExecuteMessageLoop();
			}
		}
		
		if (! ExitCode) GetExitCodeProcess(pi.hProcess, & ExitCode);
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);

		while (WaitingForUser && *WaitingForUser) {
			Sleep(10);
			ExecuteMessageLoop();
		}
	} else {
		ExitCode = 0xFFFFFFFF;
	}

	if (ErrFile) CloseHandle(ErrFile);
	if (OutFile) CloseHandle(OutFile);
	if (InFile) CloseHandle(InFile);
	if (FreeCommandLine) FatalFree(FullCommandLine);
	
	return ! ExitCode;
}

BOOL ExecutePipedCommands(LPCTSTR ApplicationName1, LPCTSTR CommandLine1, 
						  LPCTSTR ApplicationName2, LPCTSTR CommandLine2, 
                          LPCTSTR InFileName, LPCTSTR OutFileName, LPCTSTR ErrFileName,
                          LPBOOL IsCancelled, LPBOOL WaitingForUser)
{
	BOOL Result = FALSE;
	
	BOOL FreeCommandLine1 = FALSE;
	LPTSTR FullCommandLine1 = NULL;
	BOOL FreeCommandLine2 = FALSE;
	LPTSTR FullCommandLine2 = NULL;
	SECURITY_ATTRIBUTES sa;
	HANDLE InFile    = INVALID_HANDLE_VALUE;
	HANDLE ReadPipe  = INVALID_HANDLE_VALUE;
	HANDLE WritePipe = INVALID_HANDLE_VALUE;
	HANDLE OutFile   = INVALID_HANDLE_VALUE;
	HANDLE ErrFile   = INVALID_HANDLE_VALUE;
	STARTUPINFO si;
	PROCESS_INFORMATION pi1, pi2;
	DWORD ExitCode1 = 0;
	DWORD ExitCode2 = 0;

	ZeroMemory(& sa, sizeof(SECURITY_ATTRIBUTES));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.bInheritHandle = TRUE;

	if (ApplicationName1 && CommandLine1) {
		FullCommandLine1 = (LPTSTR) FatalAlloc((lstrlen(ApplicationName1) + 4 + lstrlen(CommandLine1)) * sizeof(TCHAR));
		if (! FullCommandLine1) goto EndOfFunction;
		wsprintf(FullCommandLine1, TEXT("\"%s\" %s"), ApplicationName1, CommandLine1);
		FreeCommandLine1 = TRUE;
	} else if (ApplicationName1) {
		FullCommandLine1 = (LPTSTR) ApplicationName1;
	} else {
		FullCommandLine1 = (LPTSTR) CommandLine1;
	}

	if (ApplicationName2 && CommandLine2) {
		FullCommandLine2 = (LPTSTR) FatalAlloc((lstrlen(ApplicationName2) + 4 + lstrlen(CommandLine2)) * sizeof(TCHAR));
		if (! FullCommandLine2) goto EndOfFunction;
		wsprintf(FullCommandLine2, TEXT("\"%s\" %s"), ApplicationName2, CommandLine2);
		FreeCommandLine2 = TRUE;
	} else if (ApplicationName2) {
		FullCommandLine2 = (LPTSTR) ApplicationName2;
	} else {
		FullCommandLine2 = (LPTSTR) CommandLine2;
	}

	if (InFileName) {
		InFile = CreateFile(InFileName, GENERIC_READ, FILE_SHARE_READ, 
			& sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (InFile == INVALID_HANDLE_VALUE) goto EndOfFunction;
	}
	if (! CreatePipe(& ReadPipe, & WritePipe,  & sa, 0)) {
		goto EndOfFunction;
	}
	if (OutFileName) {
		OutFile = CreateFile(OutFileName, GENERIC_WRITE, FILE_SHARE_READ, 
			& sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (OutFile == INVALID_HANDLE_VALUE) goto EndOfFunction;
	}
	if (ErrFileName) {
		ErrFile = CreateFile(ErrFileName, GENERIC_WRITE, FILE_SHARE_READ, 
			& sa, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (ErrFile == INVALID_HANDLE_VALUE) {
			goto EndOfFunction;
		} else {
			SetFilePointer(ErrFile, 0, NULL, FILE_END);
		}
	}

	ZeroMemory(& si, sizeof(STARTUPINFO));
	si.cb          = sizeof(STARTUPINFO);
	si.dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
	si.wShowWindow = SW_SHOWNORMAL;
	si.hStdInput   = ((InFile != INVALID_HANDLE_VALUE) ? InFile : GetStdHandle(STD_INPUT_HANDLE));
	si.hStdOutput  = WritePipe;
	si.hStdError   = GetStdHandle(STD_ERROR_HANDLE); // ((ErrFile != INVALID_HANDLE_VALUE) ? ErrFile : GetStdHandle(STD_ERROR_HANDLE));
	if (CreateProcess(ApplicationName1, FullCommandLine1, NULL, NULL, TRUE, 
		              CREATE_NO_WINDOW, NULL, NULL, &si, &pi1)) {
		si.hStdInput   = ReadPipe;
		si.hStdOutput  = ((OutFile != INVALID_HANDLE_VALUE) ? OutFile : GetStdHandle(STD_OUTPUT_HANDLE));
		si.hStdError   = ((ErrFile != INVALID_HANDLE_VALUE) ? ErrFile : GetStdHandle(STD_ERROR_HANDLE));
		if (CreateProcess(ApplicationName2, FullCommandLine2, NULL, NULL, TRUE, 
		                    CREATE_NO_WINDOW, NULL, NULL, &si, &pi2)) {

			while (WaitForSingleObject(pi1.hProcess, 10) == WAIT_TIMEOUT) {
				if (IsCancelled && *IsCancelled) {
					TerminateProcess(pi1.hProcess, 0xFFFFFFFF);
					ExitCode1 = 0xFFFFFFFF;
				} else {
					if (IsCancelled) ExecuteMessageLoop();
				}
			}
			if (! ExitCode1) GetExitCodeProcess(pi1.hProcess, & ExitCode1);
			CloseHandle(pi1.hProcess);
			CloseHandle(pi1.hThread);
			CloseHandle(WritePipe);
			WritePipe = INVALID_HANDLE_VALUE;

			while (WaitForSingleObject(pi2.hProcess, 10) == WAIT_TIMEOUT) {
				if (IsCancelled && *IsCancelled) {
					TerminateProcess(pi2.hProcess, 0xFFFFFFFF);
					ExitCode2 = 0xFFFFFFFF;
				} else {
					if (IsCancelled) ExecuteMessageLoop();
				}
			}
			if (! ExitCode2) GetExitCodeProcess(pi2.hProcess, & ExitCode2);
			CloseHandle(pi2.hProcess);
			CloseHandle(pi2.hThread);
			CloseHandle(ReadPipe);
			ReadPipe = INVALID_HANDLE_VALUE;

			Result = (ExitCode1 == 0 && ExitCode2 == 0);
			
		} else {
			TerminateProcess(pi1.hProcess, 0xFFFFFFFF);
			WaitForSingleObject(pi1.hProcess, INFINITE);
			CloseHandle(pi1.hProcess);
			CloseHandle(pi1.hThread);
		}
	}

EndOfFunction:;
	if (ErrFile   != INVALID_HANDLE_VALUE) CloseHandle(ErrFile);
	if (OutFile   != INVALID_HANDLE_VALUE) CloseHandle(OutFile);
	if (WritePipe != INVALID_HANDLE_VALUE) CloseHandle(WritePipe);
	if (ReadPipe  != INVALID_HANDLE_VALUE) CloseHandle(ReadPipe);
	if (InFile    != INVALID_HANDLE_VALUE) CloseHandle(InFile);
	if (FreeCommandLine2) FatalFree(FullCommandLine2);
	if (FreeCommandLine1) FatalFree(FullCommandLine1);
	
	return Result;
}

typedef struct tagGETFONTINFOPARAM {
	LPTSTR FaceName;
	LOGFONT * Font;
	TEXTMETRIC * Metric;
	BOOL Found;
} GETFONTINFOPARAM, * LPGETFONTINFOPARAM;

int CALLBACK GetFontInfoProc(CONST LOGFONT * Font, CONST TEXTMETRIC * Metric, DWORD Type, LPARAM Data)
{
	LPGETFONTINFOPARAM Param;
	Param = (LPGETFONTINFOPARAM) Data;
	if (! lstrcmpi(Param->FaceName, Font->lfFaceName)) {
		if (Param->Font) CopyMemory(Param->Font, Font, sizeof(LOGFONT));
		if (Param->Metric) CopyMemory(Param->Metric, Metric, sizeof(TEXTMETRIC));
		Param->Found = TRUE;
	}
	return 0;
}

BOOL GetFontInfo(LPCTSTR FaceName, LOGFONT * Font, TEXTMETRIC * Metric)
{
	HDC dc;
	GETFONTINFOPARAM Param;
	Param.FaceName = (LPTSTR) (LPVOID) FaceName;
	Param.Font     = Font;
	Param.Metric   = Metric;
	Param.Found    = FALSE;
	dc = GetDC(HWND_DESKTOP);
	EnumFonts(dc, FaceName, GetFontInfoProc, (LPARAM) & Param);
	ReleaseDC(HWND_DESKTOP, dc);
	return Param.Found;
}

typedef struct tagGETFONTLISTPARAM {
	LOGFONT * Font;
	int Count;
} GETFONTLISTPARAM, * LPGETFONTLISTPARAM;

int CALLBACK GetFontListProc(CONST LOGFONT * Font, CONST TEXTMETRIC * Metric, DWORD Type, LPARAM Data)
{
	LPGETFONTLISTPARAM Param;
	Param = (LPGETFONTLISTPARAM) Data;
	Param->Font = (LOGFONT *) FatalReAlloc(Param->Font, (Param->Count + 1) * sizeof(LOGFONT));
	CopyMemory(Param->Font + Param->Count, Font, sizeof(LOGFONT));
	return 1;
}

LOGFONT * GetFontList(int * Count)
{
	HDC dc;
	GETFONTLISTPARAM Param;
	Param.Font = NULL;
	Param.Count = 0;
	dc = GetDC(HWND_DESKTOP);
	EnumFonts(dc, TEXT(""), GetFontListProc, (LPARAM) & Param);
	ReleaseDC(HWND_DESKTOP, dc);
	*Count = Param.Count;
	return Param.Font;
}
