#ifndef WINUTILS_H_
#define WINUTILS_H_

#include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif

LPVOID FatalAlloc(DWORD Size);

LPVOID FatalReAlloc(LPVOID Ptr, DWORD Size);

VOID FatalFree(LPVOID Ptr);

LPTSTR CopyStr(LPTSTR Dst, LPCTSTR Src);

LPTSTR ConcatStr(LPTSTR Dst, LPCTSTR Src);

LPTSTR DoubleStr(double d);

LPTSTR GetAllocFullPathName(LPCTSTR lpFileName,	LPTSTR * lpFilePart);

/* Helper functions to create files in Windows's special folders.
*/

LPTSTR GetFullPathInSpecialFolder(
	int csidl, 
	LPCTSTR lpMainDirectory, 
	LPCTSTR lpSubDirectory, 
	LPCTSTR lpFileName, 
	LPTSTR lpFullPathName, 
	UINT uSize, 
	BOOL bCreateDirectory, 
	LPSECURITY_ATTRIBUTES lpSecurityAttributes);
/* Provides the full path to a file in a special folder of Windows. 
   Optionally, the function can create the folder in which the file shall be created.
   Parameters:
   - csidl
      Identifier of the special folder inside which the file shall be placed. 
      It shall be set to one of the CSIDL_ values defined in the Win32 API.
      If the required special folder is not supported by the actual version
      of Windows, the full path returned will be inside the application's
      main directory.
   - lpMainDirectory
      The directory which shall be created inside the special folder for the
      current application. If the required special folder is not supported by 
      the actual version of Windows and therefore the application's main 
      directory is used, this parameter is ignored. This parameter may be NULL.
   - lpSubDirectory
      The directory which shall be created inside lpMainDirectory in the 
      special folder or directly inside the application's main directory.
      This parameter may be NULL.
   - lpFileName
      The name of the file. This parameter may be NULL if only the directory
      name is required.
   - lpFullPathName
      If this parameter is not NULL, it shall point to a buffer into which
      the full path will be copied. The size of the buffer shall be given
      in uSize.
      If this parameter is NULL, the function will allocate a buffer for 
      the full path. The allocated buffer shall be released using LocalFree.
   - uSize
      If lpFullPathName is not NULL, this parameter shall contain the size
      of the buffer.
   - bCreateDirectory
      If this parameter is set to TRUE, the function will create the directory
      in the filesystem if it's not there yet.
   - lpSecurityAttributes
      To be passed to the CreateDirectory Win32 API function when directories
      are created.
   Return value:
      If the function succeeds, it returns the pointer to the full path name
      of the file in the given special folder.
      If an error occured, the return value is NULL; GetLastError() may return 
      something useful afterwards. Possible causes are:
      - The buffer passed as lpFullPathName is too small.
      - The directory/file names given as lpMainDirectory, lpSubDirectory or
        lpFileName are invalid.
      - The directory cannot be created.
*/      

LPTSTR * GetOpenFileList(LPTSTR FileNames, int * FileCount);

VOID FreeOpenFileList(LPTSTR * FileList);

BOOL BrowseForDirectory(HWND hWnd, LPTSTR Folder, LPCTSTR Title, BOOL CreateButton);

BOOL CreateDirectoryTree(
	LPCTSTR lpFullPathName, 
	BOOL bLastTokenIsFileName,                   
	LPSECURITY_ATTRIBUTES lpSecurityAttributes);
/* Something like CreateDirectory in the Win32 API except that this function
   can create directories recursively.
   Parameters:
   - lpFullPathName
      Path of the directory to be created.
   - bLastTokenIsFileName
      If this parameter is TRUE, the function will assume that the characters
      after the last (back)slash specify a file name; in this case, only the
      directory for the file will be created.
   - lpSecurityAttributes
      To be passed to the CreateDirectory Win32 API function.
   The return value is TRUE if the function succeeds and FALSE if it fails.
*/

LPTSTR ChangeToDirectory(LPCTSTR Directory);

BOOL ChangeDirectoryBack(LPTSTR OrigDirectory);

BOOL CleanDirectory(LPCTSTR Directory);

BOOL DeleteFilesFromDirectory(LPCTSTR Directory, LPCTSTR FindMask);

VOID ExecuteMessageLoop();

BOOL ExecuteCommand(LPCTSTR ApplicationName, LPCTSTR CommandLine, 
                    LPCTSTR InFileName, LPCTSTR OutFileName, LPCTSTR ErrFileName,
                    LPBOOL IsCancelled, LPBOOL WaitingForUser);

BOOL ExecutePipedCommands(LPCTSTR ApplicationName1, LPCTSTR CommandLine1, 
						  LPCTSTR ApplicationName2, LPCTSTR CommandLine2, 
                          LPCTSTR InFileName, LPCTSTR OutFileName, LPCTSTR ErrFileName,
                          LPBOOL IsCancelled, LPBOOL WaitingForUser);

BOOL GetFontInfo(LPCTSTR FontName, LOGFONT * FontInfo, TEXTMETRIC * Metric);

LOGFONT * GetFontList(int * Count);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* WINUTILS_H_ */
